import type { BackendQueryFormat, NodeStruct } from 'ts-common';
import type { QueryText } from '../model';

type Logic = any;

function cleanStringOfNewLinesAndTabs(str: string): string {
  if (!str) {
    return '';
  }
  let trimmedSQL = str.replace(/\n/g, ' ');
  trimmedSQL = trimmedSQL.replaceAll(/ {2,50}/g, ' ');
  trimmedSQL = trimmedSQL.replace(/\t+/g, '');
  return trimmedSQL.trim();
}

function buildNodeCTE(node: NodeStruct | undefined): string[] {
  if (!node || !node.label || !node.id) {
    return [];
  }
  const nodeParts: string[] = [];
  const mainNode = `
  SELECT jsonb_build_object(
    '_id', concat('${node.label}-', ${node.id}.id::text),
    'label', '${node.label}',
    'attributes', to_jsonb(${node.id}) - 'id'
  ) AS node
  FROM ${node.label} ${node.id}`;
  nodeParts.push(cleanStringOfNewLinesAndTabs(mainNode));
  buildNodeCTE(node.relation?.node).forEach(part => nodeParts.push(part));
  return nodeParts;
}

function buildEdgesCTE(node: NodeStruct | undefined): string[] {
  if (!node || !node.relation || !node.relation.node || !node.id || !node.relation.node.id || !node.relation.label) {
    return [];
  }

  const edge = `
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('${node.label}-', ${node.id}.id::text, '-${node.relation.node.label}-', ${node.relation.node.id}.id::text),
      'label', '${node.relation.label}',
      'attributes', '{}'::jsonb,
      'from', ${node.id}.id::text,
      'to', ${node.relation.node.id}.id::text
    )
  ) AS edges
  FROM ${node.label} ${node.id}
  JOIN ${node.relation.node.label} ${node.relation.node.id} ON ${node.id}.${node.relation.label.split('.')[0]} = ${node.relation.node.id}.${
    node.relation.label.split('.')[1]
  }`;
  const ret = cleanStringOfNewLinesAndTabs(edge);
  return [ret, ...buildEdgesCTE(node.relation.node)];
}

function buildNodeCountsCTE(node: NodeStruct | undefined): string[] {
  if (!node || !node.label || !node.id) {
    return [];
  }
  const nodeParts: string[] = [];
  const mainNode = `SELECT '${node.label}' AS label FROM ${node.label}`;
  nodeParts.push(cleanStringOfNewLinesAndTabs(mainNode));
  buildNodeCountsCTE(node.relation?.node).forEach(part => nodeParts.push(part));
  return nodeParts;
}

function buildEdgeCountsCTE(node: NodeStruct | undefined): string[] {
  if (!node || !node.relation || !node.relation.node || !node.id || !node.relation.node.id || !node.relation.label) {
    return [];
  }

  const edge = `
  SELECT '${node.relation.label}' AS label, COUNT(*) AS count
  FROM ${node.label} ${node.id}
  JOIN ${node.relation.node.label} ${node.relation.node.id}
  ON ${node.id}.${node.relation.label.split('.')[0]} = ${node.relation.node.id}.${node.relation.label.split('.')[1]}
  GROUP BY label`;
  const ret = cleanStringOfNewLinesAndTabs(edge);
  return [ret, ...buildEdgeCountsCTE(node.relation.node)];
}

function buildWhereClause(logic: Logic): string {
  if (Array.isArray(logic) && logic.length === 3) {
    const [operator, field, value] = logic;
    const tableName = field.split('.')[0];
    const columnName = field.split('.')[1];
    return `${tableName}.${columnName} ${operator} ${value}`;
  }
  return '';
}

export function query2SQL(query: BackendQueryFormat): QueryText {
  const nodesCTE = [...new Set(query.query.flatMap(path => buildNodeCTE(path.node)))];
  const edgesCTE = [...new Set(query.query.flatMap(path => buildEdgesCTE(path.node)))];
  const nodeCountsCTE = [...new Set(query.query.flatMap(path => buildNodeCountsCTE(path.node)))];
  const edgeCountsCTE = [...new Set(query.query.flatMap(path => buildEdgeCountsCTE(path.node)))];
  const whereClause = buildWhereClause(query.logic);

  return {
    query: `WITH nodes_cte AS (
      SELECT jsonb_agg(node) AS nodes
      FROM (
        ${nodesCTE.join(' UNION ALL ')}
      ) sub
    ),
    edges_cte AS (
      ${edgesCTE.join(' UNION ALL ')}
    ),
    node_counts AS (
      SELECT label, COUNT(*) AS count
      FROM (
        ${nodeCountsCTE.join(' UNION ALL ')}
      ) t
      GROUP BY label
    ),
    edge_counts AS (
      ${edgeCountsCTE.join(' UNION ALL ')}
    )
    SELECT jsonb_build_object(
      'nodes', nodes_cte.nodes,
      'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
      'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                               FROM node_counts), '[]'::jsonb),
      'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                               FROM edge_counts), '[]'::jsonb)
    ) AS result FROM nodes_cte, edges_cte
    ${whereClause ? `WHERE ${whereClause}` : ''}
    ${query.limit ? `LIMIT ${query.limit}` : ''};`,
    countQuery: '',
  };
}
