import { log } from '../../logger';
import type { CountQueryResultFromBackend, EdgeQueryResult, NodeQueryResult } from 'ts-common';
import type { GraphQueryResultFromBackend } from 'ts-common';
import { type QueryResult } from 'pg';

// Adjusted to handle a Postgres QueryResult object.
export function parseSQLQueryResult(result: QueryResult, returnType: 'nodelink' | 'table' = 'nodelink'): GraphQueryResultFromBackend {
  // ...existing error handling code...
  try {
    // Extract rows if present.
    switch (returnType) {
      case 'nodelink':
        return parseSQLNodeLinkQuery(result.rows);
      case 'table':
        log.error(`Table format not supported yet`);
        throw new Error('Table format not supported yet');
      default:
        log.error(`Error Unknown query Format`);
        throw new Error('Unknown query Format');
    }
  } catch (err) {
    log.error(`Error executing query`, err);
    throw err;
  }
}

// Adjusted to handle a Postgres QueryResult object.
export function parseCountSQLQueryResult(result: QueryResult): CountQueryResultFromBackend {
  try {
    const countResult: CountQueryResultFromBackend = { updatedAt: Date.now() };
    const rows = result.rows as Record<string, any>[];
    if (rows.length > 0) {
      const row = rows[0];
      for (const key in row) {
        if (Object.prototype.hasOwnProperty.call(row, key)) {
          countResult[key] = Number(row[key]);
        }
      }
    }
    return countResult;
  } catch (err) {
    log.error(`Error executing query`, err);
    throw err;
  }
}

// Helper function to build a graph result (nodelink) from SQL rows.
function parseSQLNodeLinkQuery(rows: Record<string, any>[]): GraphQueryResultFromBackend {
  const nodes: NodeQueryResult[] = [];
  const edges: EdgeQueryResult[] = [];
  const seenNodes = new Set<string>();
  const seenEdges = new Set<string>();

  for (const row of rows) {
    // If the row doesn't carry a "type", assume it represents a node.
    if (!row.type) {
      if (!seenNodes.has(row.id)) {
        nodes.push({
          _id: row.id,
          label: row.name, // using "name" as the label
          attributes: { ...row },
        });
        seenNodes.add(row.id);
      }
    } else if (row.type === 'node') {
      if (!seenNodes.has(row.id)) {
        nodes.push({
          _id: row.id,
          label: row.label,
          attributes: row.attributes, // Assumed to be a plain object.
        });
        seenNodes.add(row.id);
      }
    } else if (row.type === 'edge') {
      if (!seenEdges.has(row.id)) {
        edges.push({
          _id: row.id,
          label: row.label,
          from: row.from,
          to: row.to,
          attributes: { ...row.attributes, type: row.label },
        });
        seenEdges.add(row.id);
      }
    } else if (row.type === 'path') {
      // If a path, assume arrays "nodes" and "edges" exist.
      if (Array.isArray(row.nodes)) {
        for (const node of row.nodes) {
          if (!seenNodes.has(node.id)) {
            nodes.push({
              _id: node.id,
              label: node.label,
              attributes: node.attributes,
            });
            seenNodes.add(node.id);
          }
        }
      }
      if (Array.isArray(row.edges)) {
        for (const edge of row.edges) {
          if (!seenEdges.has(edge.id)) {
            edges.push({
              _id: edge.id,
              label: edge.label,
              from: edge.from,
              to: edge.to,
              attributes: { ...edge.attributes, type: edge.label },
            });
            seenEdges.add(edge.id);
          }
        }
      }
    } else {
      log.warn(`Ignoring unknown row type: ${row.type}`);
    }
  }

  return { nodes, edges };
}

// Removed neo4j-specific helper functions.
