import { type DbConnection, type GraphQueryResultMetaFromBackend, graphQueryBackend2graphQuery } from 'ts-common';
import { PgConnection } from 'ts-common/databaseConnection/postgres';
import { log } from '../../logger';
import { QUERY_CACHE_DURATION, redis } from '../../variables';
import { parseCountCypherQueryResult, parseCypherQueryResult } from '../../queryExecution/cypher/queryResultParser';
import { cacheCheck } from './cache';
import type { QueryText } from '../../queryExecution/model';
import { parseCountSQLQueryResult, parseSQLQueryResult } from '../../queryExecution/sql/queryResultParser';

export const sqlQueryService = async (
  db: DbConnection,
  queryText: QueryText,
  useCached: boolean,
): Promise<GraphQueryResultMetaFromBackend> => {
  let index = 0;
  const disambiguatedQuery = queryText.query.replace(/\d{13}/g, () => (index++).toString());
  const cacheKey = Bun.hash(JSON.stringify({ db: db, query: disambiguatedQuery })).toString();

  // if (QUERY_CACHE_DURATION === '') {
  //   log.info('Query cache disabled, skipping cache check');
  // } else if (!useCached) {
  //   log.info('Skipping cache check for query due to parameter', useCached);
  // } else {
  //   const cachedMessage = await cacheCheck(cacheKey);
  //   if (cachedMessage) {
  //     log.debug('Cache hit for query', disambiguatedQuery);
  //     return cachedMessage;
  //   }
  // }

  const connection = new PgConnection(db);
  try {
    const [result, countResult] = await connection.run([queryText.query, queryText.countQuery]);
    // console.log('result:', result);
    // console.log('countResult:', countResult);

    const graph = parseSQLQueryResult(result);
    log.info('Parsed graph:', result);
    const countGraph = parseCountSQLQueryResult(countResult);

    // calculate metadata
    const graphQueryResult = graphQueryBackend2graphQuery(graph, countGraph);
    graphQueryResult.nodeCounts.updatedAt = Date.now();

    // cache result
    const compressedMessage = Bun.gzipSync(JSON.stringify(result));
    const base64Message = Buffer.from(compressedMessage).toString('base64');

    if (QUERY_CACHE_DURATION !== '') {
      // if cache enabled, cache the result
      await redis.setWithExpire(cacheKey, base64Message, QUERY_CACHE_DURATION); // ttl in seconds
    }

    // log.info('Query result:', graphQueryResult);

    return graphQueryResult;
  } catch (error) {
    log.error('Error parsing query result:', queryText, error);
    throw new Error('Error parsing query result');
  } finally {
    connection.close();
  }
};
